/*
 * Copyright (c) 2021 Cloudera, Inc. All rights reserved.
 *
 * This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 * Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 * to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 * properly licensed third party, you do not have any rights to this code.
 *
 * If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 */

package com.cloudera.kafka.cruisecontrol.detector;

import com.linkedin.cruisecontrol.detector.metricanomaly.MetricAnomaly;
import com.linkedin.cruisecontrol.detector.metricanomaly.MetricAnomalyFinder;
import com.linkedin.cruisecontrol.detector.metricanomaly.MetricAnomalyType;
import com.linkedin.cruisecontrol.exception.NotEnoughValidWindowsException;
import com.linkedin.cruisecontrol.monitor.sampling.aggregator.ValuesAndExtrapolations;
import com.linkedin.kafka.cruisecontrol.KafkaCruiseControl;
import com.linkedin.kafka.cruisecontrol.async.progress.OperationProgress;
import com.linkedin.kafka.cruisecontrol.config.KafkaCruiseControlConfig;
import com.linkedin.kafka.cruisecontrol.config.KafkaCruiseControlConfigUtils;
import com.linkedin.kafka.cruisecontrol.config.constants.AnomalyDetectorConfig;
import com.linkedin.kafka.cruisecontrol.config.constants.MonitorConfig;
import com.linkedin.kafka.cruisecontrol.exception.KafkaCruiseControlException;
import com.linkedin.kafka.cruisecontrol.model.ClusterModel;
import com.linkedin.kafka.cruisecontrol.monitor.ModelCompletenessRequirements;
import com.linkedin.kafka.cruisecontrol.monitor.sampling.holder.BrokerEntity;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import static com.linkedin.kafka.cruisecontrol.analyzer.goals.GoalUtils.MIN_NUM_VALID_WINDOWS_FOR_SELF_HEALING;
import static com.linkedin.kafka.cruisecontrol.config.constants.AnomalyDetectorConfig.SELF_HEALING_EXCLUDE_RECENTLY_DEMOTED_BROKERS_CONFIG;
import static com.linkedin.kafka.cruisecontrol.config.constants.AnomalyDetectorConfig.SELF_HEALING_EXCLUDE_RECENTLY_REMOVED_BROKERS_CONFIG;
import static com.linkedin.kafka.cruisecontrol.detector.AnomalyDetectorUtils.ANOMALY_DETECTION_TIME_MS_OBJECT_CONFIG;
import static com.linkedin.kafka.cruisecontrol.detector.AnomalyDetectorUtils.KAFKA_CRUISE_CONTROL_OBJECT_CONFIG;
import static com.linkedin.kafka.cruisecontrol.detector.MetricAnomalyDetector.METRIC_ANOMALY_BROKER_ENTITIES_OBJECT_CONFIG;
import static com.linkedin.kafka.cruisecontrol.detector.MetricAnomalyDetector.METRIC_ANOMALY_FIXABLE_OBJECT_CONFIG;


/**
 * This class will check whether there is empty broker from collected broker metrics.
 * Empty brokers are identified by checking one metric for each broker - replica count.
 */
public class EmptyBrokerAnomalyFinder implements MetricAnomalyFinder<BrokerEntity> {

    private static final Logger LOG = LoggerFactory.getLogger(EmptyBrokerAnomalyFinder.class);

    private KafkaCruiseControl _kafkaCruiseControl;
    private final Map<MetricAnomalyType, Integer> _numEmptyBrokersByType;

    public EmptyBrokerAnomalyFinder() {
        _numEmptyBrokersByType = new HashMap<>(MetricAnomalyType.cachedValues().size());
        MetricAnomalyType.cachedValues().forEach(type -> _numEmptyBrokersByType.put(type, 0));
    }

    @Override
    public Collection<MetricAnomaly<BrokerEntity>> metricAnomalies(Map<BrokerEntity, ValuesAndExtrapolations> metricsHistoryByBroker,
                                                                   Map<BrokerEntity, ValuesAndExtrapolations> currentMetricsByBroker) {
        LOG.info("Empty broker detection started.");
        try {
            Set<BrokerEntity> detectedMetricAnomalies = detectEmptyBrokers();
            Map<BrokerEntity, Long> detectedEmptyBrokers = createDetectedEmptyBrokerMap(detectedMetricAnomalies);

            _numEmptyBrokersByType.put(MetricAnomalyType.RECENT, detectedEmptyBrokers.size());
            return createEmptyBrokerAnomalies(detectedEmptyBrokers);
        } catch (Exception e) {
            LOG.warn("Empty broker detector encountered exception: ", e);
        } finally {
            LOG.info("Empty broker detection finished.");
        }
        return Collections.emptySet();
    }

    @Override
    public int numAnomaliesOfType(MetricAnomalyType type) {
        return _numEmptyBrokersByType.get(type);
    }

    private Set<BrokerEntity> detectEmptyBrokers() {
        try {
            KafkaCruiseControlConfig config = _kafkaCruiseControl.config();
            boolean allowCapacityEstimation = config.getBoolean(AnomalyDetectorConfig.ANOMALY_DETECTION_ALLOW_CAPACITY_ESTIMATION_CONFIG);
            double minValidPartitionsRatio = config.getDouble(MonitorConfig.MIN_VALID_PARTITION_RATIO_CONFIG);
            ModelCompletenessRequirements modelCompletenessRequirements = new ModelCompletenessRequirements(MIN_NUM_VALID_WINDOWS_FOR_SELF_HEALING,
                    minValidPartitionsRatio, false);

            ClusterModel clusterModel = _kafkaCruiseControl.clusterModel(modelCompletenessRequirements,
                    allowCapacityEstimation, new OperationProgress());
            Set<BrokerEntity> detectedMetricAnomalies = clusterModel.aliveBrokers().stream()
                    .filter(b -> b.replicas().isEmpty()).map(b -> new BrokerEntity(b.host().name(), b.id())).collect(Collectors.toSet());
            return getFixableEmptyBrokers(detectedMetricAnomalies);
        } catch (NotEnoughValidWindowsException nevwe) {
            LOG.debug("Skipping empty broker detection because there are not enough valid windows.", nevwe);
        } catch (KafkaCruiseControlException kcce) {
            LOG.warn("Empty broker detector received exception", kcce);
        } catch (Exception e) {
            LOG.error("Unexpected exception", e);
        }

        return new HashSet<>();
    }

    /**
     * Get the remaining empty brokers after the configured exclusions.
     *
     * @param detectedMetricAnomalies Detected empty brokers with no exclusion.
     * @return The empty brokers after the applied exclusions.
     */
    private Set<BrokerEntity> getFixableEmptyBrokers(Set<BrokerEntity> detectedMetricAnomalies) {
        KafkaCruiseControlConfig config = _kafkaCruiseControl.config();
        boolean excludeRecentlyDemotedBrokers = config.getBoolean(SELF_HEALING_EXCLUDE_RECENTLY_DEMOTED_BROKERS_CONFIG);
        boolean excludeRecentlyRemovedBrokers = config.getBoolean(SELF_HEALING_EXCLUDE_RECENTLY_REMOVED_BROKERS_CONFIG);

        Set<Integer> emptyBrokerIds = detectedMetricAnomalies.stream().mapToInt(BrokerEntity::brokerId).boxed().collect(Collectors.toSet());
        Set<Integer> recentlyDemotedBrokerIds = _kafkaCruiseControl.recentBrokers(false);
        Set<Integer> recentlyRemovedBrokerIds = _kafkaCruiseControl.recentBrokers(true);

        if (excludeRecentlyDemotedBrokers) {
            emptyBrokerIds.removeAll(recentlyDemotedBrokerIds);
        }

        if (excludeRecentlyRemovedBrokers) {
            emptyBrokerIds.removeAll(recentlyRemovedBrokerIds);
        }

        return detectedMetricAnomalies.stream().filter(a -> emptyBrokerIds.contains(a.brokerId())).collect(Collectors.toSet());
    }

    private Map<BrokerEntity, Long> createDetectedEmptyBrokerMap(Set<BrokerEntity> detectedMetricAnomalies) {
        Map<BrokerEntity, Long> detectedEmptyBrokers = new HashMap<>();

        for (BrokerEntity broker : detectedMetricAnomalies) {
            detectedEmptyBrokers.put(broker, _kafkaCruiseControl.timeMs());
        }

        return detectedEmptyBrokers;
    }

    private Set<MetricAnomaly<BrokerEntity>> createEmptyBrokerAnomalies(Map<BrokerEntity, Long> detectedEmptyBrokers) {
        Set<MetricAnomaly<BrokerEntity>> detectedEmptyBrokerSet = new HashSet<>();

        if (!detectedEmptyBrokers.isEmpty()) {
            detectedEmptyBrokerSet.add(createEmptyBrokersAnomaly(detectedEmptyBrokers));
        }

        return detectedEmptyBrokerSet;
    }

    private EmptyBrokers createEmptyBrokersAnomaly(Map<BrokerEntity, Long> detectedBrokers) {
        Map<String, Object> parameterConfigOverrides = new HashMap<>(4);

        parameterConfigOverrides.put(KAFKA_CRUISE_CONTROL_OBJECT_CONFIG, _kafkaCruiseControl);
        parameterConfigOverrides.put(METRIC_ANOMALY_BROKER_ENTITIES_OBJECT_CONFIG, detectedBrokers);
        parameterConfigOverrides.put(ANOMALY_DETECTION_TIME_MS_OBJECT_CONFIG, _kafkaCruiseControl.timeMs());
        parameterConfigOverrides.put(METRIC_ANOMALY_FIXABLE_OBJECT_CONFIG, !detectedBrokers.isEmpty());

        Map<String, Object> configPairs = _kafkaCruiseControl.config().mergedConfigValues();
        configPairs.putAll(parameterConfigOverrides);
        return KafkaCruiseControlConfigUtils.getConfiguredInstance(EmptyBrokers.class,
                EmptyBrokers.class,
                configPairs);
    }

    @Override
    public void configure(Map<String, ?> configs) {
        _kafkaCruiseControl = (KafkaCruiseControl) configs.get(KAFKA_CRUISE_CONTROL_OBJECT_CONFIG);

        if (_kafkaCruiseControl == null) {
            throw new IllegalArgumentException("Empty broker detector is missing " + KAFKA_CRUISE_CONTROL_OBJECT_CONFIG);
        }
    }
}
